// application.vala
//
// Copyright (C) 2019 Vivien Kraus <vivien@planete-kraus.eu>
// 
//  This program is free software: you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see
//  <http://www.gnu.org/licenses/>.

namespace Tarot {
	public class Application: GLib.Object {
		private Tarot.History<Tarot.Game> m_history = new Tarot.History<Tarot.Game> ();

		public int history_capacity {
			get {
				return m_history.capacity;
			}
			set {
				m_history.capacity = value;
			}
		}

		public Tarot.Game game {
			owned get {
				return m_history.current.dup ();
			}
		}

		public string xml {
			owned get {
				return m_history.current.save_to_xml_alloc ();
			}
		}

		public signal void broadcast_event (Tarot.GameEvent e);
		public signal void unicast_event (Tarot.GameEvent e, int who);

		private void added_event (Tarot.GameEvent e) {
			switch (e.type ()) {
			case Tarot.GameEventT.DEAL_EVENT:
				Tarot.Player whom;
				Tarot.Card[] cards;
				if (e.get_deal_alloc (out whom, out cards) != Tarot.GameEventError.OK) {
					assert (false);
				}
				unicast_event (e, (int) whom);
				break;
			case Tarot.GameEventT.DEAL_ALL_EVENT:
				Tarot.Player[] owners;
				if (e.get_deal_all_alloc (out owners) != Tarot.GameEventError.OK) {
					assert (false);
				}
				var n_owners = owners.length;
				assert (n_owners == 78);
				size_t n_players = m_history.current.n_players ();
				for (Tarot.Player p = 0; p < n_players; p++) {
					size_t n = 0;
					for (size_t i = 0; i < n_owners; i++) {
						if (owners[i] == p) {
							n++;
						}
					}
					assert (n != 0);
					var my_cards = new Tarot.Card[n];
					size_t i = 0;
					for (Tarot.Card c = 0; c < n_owners; c++) {
						if (owners[c] == p) {
							my_cards[i++] = c;
						}
					}
					assert (i == n);
					var personalized = new Tarot.GameEvent.alloc_deal (p, my_cards);
					unicast_event (personalized, (int) p);
				}
				break;
			case Tarot.GameEventT.DISCARD_EVENT:
				var game = m_history.current;
				size_t n_players = game.n_players ();
				Tarot.Player taker;
				bool has_taker = (game.get_taker (out taker) == Tarot.GameError.OK);
				assert (has_taker);
				for (Tarot.Player p = 0; p < n_players; p++) {
					Tarot.Card[] cards;
					if (p == taker) {
						if (game.get_full_discard_alloc (out cards) != Tarot.GameError.OK) {
							assert (false);
						}
					}
					else {
						if (game.get_public_discard_alloc (out cards) != Tarot.GameError.OK) {
							assert (false);
						}
					}
					var personalized = new Tarot.GameEvent.alloc_discard (cards);
					unicast_event (personalized, (int) p);
				}
				break;
			default:
				broadcast_event (e);
				break;
			}
		}

		public signal void reset ();

		public void send_events () {
			reset ();
			foreach (var e in m_history.current) {
				added_event (e);
			}
		}

		public bool load_from_xml (string str) {
			var game = new Tarot.Game.alloc ();
			if (game.load_from_xml (str) == 0) {
				m_history.push (game);
				send_events ();
				return true;
			} else {
				return false;
			}
		}

		public bool impute (string seed, int retry) {
			for (int i = 0; i <= retry; i++) {
				var game = m_history.current;
				if (game == null) {
					return false;
				}
				string current_seed = "%s (%d)".printf (seed, retry);
				char[] seed_bytes = new char[current_seed.length];
				for (long j = 0; j < current_seed.length; j++) {
					seed_bytes[j] = current_seed[j];
				}
				var counter = new Tarot.Counter.alloc ();
				counter.load_from_game (game);
				if (counter.and_game_impute (game, seed_bytes) == 0) {
					m_history.push (game);
					send_events ();
					return true;
				}
			}
			return false;
		}

		public bool mcts (MctsParams mp) {
			var game = m_history.current;
			if (game == null) {
				stderr.printf ("No game.\n");
				return false;
			}
			string current_seed = mp.seed;
			char[] seed_bytes = new char[current_seed.length];
			for (long j = 0; j < current_seed.length; j++) {
				seed_bytes[j] = current_seed[j];
			}
			var mcts = new Tarot.Mcts.alloc (2 * 1024 * 1024);
			if (mcts.set_base (game) != 0) {
				stderr.printf ("Invalid base.\n");
				return false;
			}
			mcts.seed (seed_bytes);
			if (mp.parameter >= 0) {
				mcts.set_parameter (mp.parameter);
			}
			if (mp.simulation_agreed >= 0) {
				mcts.set_simulation_agreed (mp.simulation_agreed);
			}
			if (mp.simulation_random >= 0) {
				mcts.set_simulation_random (mp.simulation_random);
			}
			if (mcts.run (mp.n_iterations) != 0) {
				stderr.printf ("Could not run.\n");
				return false;
			}
			var e = mcts.get_best ();
			assert (e != null);
			if (game.add_event (e) != 0) {
				stderr.printf ("Invalid outcome.\n");
				return false;
			}
			m_history.push (game);
			added_event (e);
			autoreveal (game, true);
			return true;
		}

		public bool deal_random (string seed, int retry) {
			for (int i = 0; i <= retry; i++) {
				var game = m_history.current;
				if (game == null) {
					return false;
				}
				string current_seed = "%s (%d)".printf (seed, retry);
				char[] seed_bytes = new char[current_seed.length];
				for (long j = 0; j < current_seed.length; j++) {
					seed_bytes[j] = current_seed[j];
				}
				var e = new Tarot.GameEvent.alloc_deal_all_random (game.n_players (), seed_bytes);
				if (game.add_event (e) == Tarot.GameError.OK) {
					m_history.push (game);
					added_event (e);
					return true;
				}
			}
			return false;
		}

		private bool autoreveal (Tarot.Game game, bool dispatch) {
			Tarot.Card[] dog;
			if (game.can_autoreveal_alloc (out dog)) {
				var e = new Tarot.GameEvent.alloc_dog (dog);
				if (game.add_event (e) != Tarot.GameError.OK) {
					assert (false);
				}
				if (dispatch) {
					broadcast_event (e);
					m_history.push (game);
				}
				return true;
			}
			return false;
		}

		public bool generate (size_t n_players, bool with_call, MctsParams mp, int retry) {
			var game = new Tarot.Game.alloc ();
			var e = new Tarot.GameEvent.alloc_setup (n_players, with_call);
			char[] seed_bytes = new char[mp.seed.length];
			for (long i = 0; i < mp.seed.length; i++) {
				seed_bytes[i] = mp.seed[i];
			}
			if (game.add_event (e) != 0) {
				assert (false);
			}
			for (int i = 0; i <= retry; i++) {
				string deal_current_seed = "%s (%d)".printf (mp.seed, retry);
				char[] deal_seed_bytes = new char[deal_current_seed.length];
				for (long j = 0; j < deal_current_seed.length; j++) {
					deal_seed_bytes[j] = deal_current_seed[j];
				}
				e = new Tarot.GameEvent.alloc_deal_all_random (game.n_players (), deal_seed_bytes);
				if (game.add_event (e) == 0) {
					var mcts = new Tarot.Mcts.alloc (2 * 1024 * 1024);
					if (mp.parameter >= 0) {
						mcts.set_parameter (mp.parameter);
					}
					if (mp.simulation_agreed >= 0) {
						mcts.set_simulation_agreed (mp.simulation_agreed);
					}
					if (mp.simulation_random >= 0) {
						mcts.set_simulation_random (mp.simulation_random);
					}
					autoreveal (game, false);
					while (game.step () != Tarot.Step.END) {
						if (mcts.set_base (game) != 0) {
							stderr.printf (_ ("Could not set up the MCTS.\n"));
							GLib.Process.abort ();
						}
						mcts.seed (seed_bytes);
						if (mcts.run (mp.n_iterations) != 0) {
							stderr.printf (_ ("Could not explore this game.\n"));
							GLib.Process.abort ();
						}
						e = mcts.get_best ();
						if (e == null) {
							stderr.printf (_ ("No exploration has been performed.\n"));
							GLib.Process.abort ();
						}
						if (game.add_event (e) != 0) {
							stderr.printf (_ ("ERROR: the MCTS is not coherent.\n"));
							GLib.Process.abort ();
						}
						autoreveal (game, false);
					}
					m_history.push (game);
					reset ();
					send_events ();
					return true;
				}
			}
			return false;
		}

		public string? mcts_tree_to_xml (MctsParams mp) {
			var game = m_history.current;
			if (game == null) {
				return null;
			}
			var mcts = new Tarot.Mcts.alloc (2 * 1024 * 1024);
			if (mp.parameter >= 0) {
				mcts.set_parameter (mp.parameter);
			}
			if (mp.simulation_agreed >= 0) {
				mcts.set_simulation_agreed (mp.simulation_agreed);
			}
			if (mp.simulation_random >= 0) {
				mcts.set_simulation_random (mp.simulation_random);
			}
			if (mcts.set_base (game) != 0) {
				return null;
			}
			char[] seed_bytes = new char[mp.seed.length];
			for (long i = 0; i < mp.seed.length; i++) {
				seed_bytes[i] = mp.seed[i];
			}
			mcts.seed (seed_bytes);
			if (mcts.run (mp.n_iterations) != 0) {
				return null;
			}
			return mcts.save_to_xml_alloc ();
		}

		public bool new_game (size_t n_players, bool with_call) {
			var game = new Tarot.Game.alloc ();
			var e = new Tarot.GameEvent.alloc_setup (n_players, with_call);
			if (game.add_event (e) != 0) {
				return false;
			}
			m_history.push (game);
			reset ();
			added_event (e);
			return true;
		}

		public bool deal_to (Tarot.Player whom, Tarot.Card[] cards) {
			var game = m_history.current;
			if (game == null) {
				return false;
			}
			var e = new Tarot.GameEvent.alloc_deal (whom, cards);
			if (game.add_event (e) != 0) {
				return false;
			}
			m_history.push (game);
			added_event (e);
			return true;
		}

		public bool deal_all (Tarot.Player[] owners) {
			var game = m_history.current;
			if (game == null) {
				return false;
			}
			var e = new Tarot.GameEvent.alloc_deal_all (owners);
			if (game.add_event (e) != 0) {
				return false;
			}
			m_history.push (game);
			added_event (e);
			return true;
		}

		public bool bid (Tarot.Bid bid) {
			var game = m_history.current;
			if (game == null) {
				return false;
			}
			var e = new Tarot.GameEvent.alloc_bid (bid);
			if (game.add_event (e) != 0) {
				return false;
			}
			m_history.push (game);
			added_event (e);
			return true;
		}

		public bool decl (bool decl) {
			var game = m_history.current;
			if (game == null) {
				return false;
			}
			var e = new Tarot.GameEvent.alloc_decl (decl);
			if (game.add_event (e) != 0) {
				return false;
			}
			m_history.push (game);
			added_event (e);
			autoreveal (game, true);
			return true;
		}

		public bool call (Tarot.Card call) {
			var game = m_history.current;
			if (game == null) {
				return false;
			}
			var e = new Tarot.GameEvent.alloc_call (call);
			if (game.add_event (e) != 0) {
				return false;
			}
			m_history.push (game);
			added_event (e);
			autoreveal (game, true);
			return true;
		}

		public bool dog (Tarot.Card[] dog) {
			var game = m_history.current;
			if (game == null) {
				return false;
			}
			var e = new Tarot.GameEvent.alloc_dog (dog);
			if (game.add_event (e) != 0) {
				return false;
			}
			m_history.push (game);
			added_event (e);
			return true;
		}

		public bool discard (Tarot.Card[] discard) {
			var game = m_history.current;
			if (game == null) {
				return false;
			}
			var e = new Tarot.GameEvent.alloc_discard (discard);
			if (game.add_event (e) != 0) {
				return false;
			}
			m_history.push (game);
			added_event (e);
			return true;
		}

		public bool handful (Tarot.Card[] handful) {
			var game = m_history.current;
			if (game == null) {
				return false;
			}
			var e = new Tarot.GameEvent.alloc_handful (handful);
			if (game.add_event (e) != 0) {
				return false;
			}
			m_history.push (game);
			added_event (e);
			return true;
		}

		public bool play (Tarot.Card play) {
			var game = m_history.current;
			if (game == null) {
				return false;
			}
			var e = new Tarot.GameEvent.alloc_card (play);
			if (game.add_event (e) != 0) {
				return false;
			}
			m_history.push (game);
			added_event (e);
			return true;
		}

		public Application () {
			m_history.push (new Tarot.Game.alloc ());
		}
	}
}