// file card_painter.vala
//
// Copyright (C) 2017, 2018 Vivien Kraus <vivien@planete-kraus.eu>
//
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see
// <http://www.gnu.org/licenses/>.

using Cairo;

namespace TarotCairo {
	public interface CardPainter: GLib.Object {

		class Reference {
			public size_t index;

			public Reference (size_t i) {
				index = i;
			}
		}

		/**
		 * Paint a card to the cairo context.  Usually this is simply
		 * printing a pixmap of width given by the property
		 * card_width.  The cards from 78 and up should be painted as
		 * a card back.
		 */
		public abstract void paint_card (Tarot.Card card, bool is_selected, Cairo.Context context);
		public abstract double card_width { get; }
		public abstract double card_height { get; }

		private void paint_positions (Tarot.Card[] cards, bool[] is_selected, Tarot.Position[] positions, Cairo.Context context) {
			assert (cards.length == is_selected.length);
			assert (cards.length == positions.length);
			var order = new Reference[cards.length];
			for (size_t i = 0; i < cards.length; i++) {
				order[i] = new Reference (i);
			}
			GLib.qsort_with_data<Reference> (order, sizeof (Reference), (a, b) => {
					size_t ia = a.index;
					size_t ib = b.index;
					Tarot.Position pa = positions[ia];
					Tarot.Position pb = positions[ib];
					double za = pa.z;
					double zb = pb.z;
					if (za < zb) {
						return -1;
					}
					return 1;
				});
			for (int i = 0; i < cards.length; i++) {
				var p = positions[order[i].index];
				var c = cards[order[i].index];
				var b = is_selected[order[i].index];
				if (c <= 78) {
					context.save ();
					context.translate (p.x, p.y);
					context.rotate (p.angle);
					context.scale (p.scale, p.scale);
					paint_card (c, b, context);
					context.restore ();
				}
			}
		}

		/* Cards above 78 are ignored */
		public void paint_center (double view_w, double view_h, Tarot.Card[] cards, bool[] is_selected, Cairo.Context context) {
			Tarot.Position[] positions;
			Tarot.layout_center_alloc (view_w, view_h, card_width, card_height,
									   cards.length, out positions);
			paint_positions (cards, is_selected, positions, context);
		}
		
		public void paint_hand (double view_w, double view_h, Tarot.Card[] cards, bool[] is_selected, Cairo.Context context) {
			Tarot.Position[] positions;
			Tarot.layout_hand_alloc (view_w, view_h, card_width, card_height,
									 cards.length, out positions);
			paint_positions (cards, is_selected, positions, context);
		}
		
		public void paint_trick (double view_w, double view_h, Tarot.Card[] cards, bool[] is_selected, Cairo.Context context) {
			Tarot.Position[] positions;
			Tarot.layout_trick_alloc (view_w, view_h, card_width, card_height,
									  cards.length, out positions);
			paint_positions (cards, is_selected, positions, context);
		}

		public void paint_game_as (double view_w,
								   double view_h,
								   Tarot.Game game,
								   Tarot.Player view,
								   Cairo.Context context) {
			size_t n_players = game.n_players ();
			var my_cards = new Tarot.Card[0];
			var cards_trick = new Tarot.Card[n_players];
			size_t current_trick;
			bool has_current_trick = (game.get_current_trick (out current_trick) == Tarot.GameError.OK);
			Tarot.Card[] foreground = {};
			var trick_content = new Tarot.GameCardAnalysis[0];
			bool has_trick_content = (has_current_trick
									  && (game.get_trick_cards_alloc (current_trick,
																	  out trick_content)
										  == Tarot.GameError.OK));
			if (has_trick_content && current_trick > 0 && trick_content.length == 0) {
				// uninteresting empty trick
				current_trick--;
				has_trick_content = (has_current_trick
									 && (game.get_trick_cards_alloc (current_trick,
																	 out trick_content)
										 == Tarot.GameError.OK));
			}
			Tarot.Player trick_leader;
			bool has_trick_leader = has_current_trick;
			if (has_trick_leader) {
				has_trick_leader = (game.get_trick_leader (current_trick, out trick_leader)
									== Tarot.GameError.OK);
			} else {
				// Unreachable, but prevent use of possibly unassigned
				// local variable
				trick_leader = 42;
			}
			assert (has_trick_leader == has_current_trick);
			for (size_t i = 0; i < n_players; i++) {
				cards_trick[i] = 79; // Not painted
			}
			if (game.get_cards_alloc (view, out my_cards) != Tarot.GameError.OK) {
				// Fill with backs
				size_t n = my_cards.length;
				for (int i = 0; i < n; i++) {
					my_cards[i] = 78;
				}
			}
			var card_of = new Tarot.Card[n_players];
			for (int i = 0; i < n_players; i++) {
				card_of[i] = 79;
			}
			if (has_trick_leader) {
				assert (has_trick_content);
				assert (trick_content.length <= n_players);
				for (int i = 0; i < trick_content.length; i++) {
					var p = (trick_leader + i) % n_players;
					card_of[p] = trick_content[i].card;
				}
			}
			for (int i = 0; i < n_players; i++) {
				Tarot.Player p = (Tarot.Player) (((size_t) (view + i)) % n_players);
				cards_trick[i] = card_of[p];
			}
			if (game.step () == Tarot.Step.DISCARD) {
				if (game.get_dog_alloc (out foreground) != Tarot.GameError.OK) {
					assert (false);
				}
			} else {
				size_t i_trick;
				Tarot.GameCardAnalysis[] trick_cards = {};
				if (game.get_current_trick (out i_trick) == Tarot.GameError.OK
					&& i_trick == 0
					&& game.get_trick_cards_alloc (i_trick, out trick_cards) == Tarot.GameError.OK
					&& trick_cards.length == 0
					&& game.get_public_discard_alloc (out foreground) == Tarot.GameError.OK) {
					// foreground is set to the public discard
				}
			}
			if (foreground.length == 0) {
				// Print the trick up, and the hand down
				var selected = new bool[n_players];
				for (size_t i = 0; i < n_players; i++) {
					selected[i] = false;
				}
				context.save ();
				context.translate (0, view_h / 4.0);
				paint_trick (view_w, view_h / 2.0, cards_trick, selected, context);
				context.restore ();
				context.save ();
				selected = new bool[my_cards.length];
				for (size_t i = 0; i < my_cards.length; i++) {
					selected[i] = false;
				}
				context.translate (0, -view_h / 4.0);
				paint_hand (view_w, view_h / 2.0, my_cards, selected, context);
				context.restore ();
			} else {
				var selected = new bool[n_players];
				for (size_t i = 0; i < n_players; i++) {
					selected[i] = false;
				}
				context.save ();
				context.translate (0, view_h / 3.0);
				paint_trick (view_w, view_h / 3.0, cards_trick, selected, context);
				context.restore ();
				selected = new bool[foreground.length];
				for (size_t i = 0; i < foreground.length; i++) {
					selected[i] = false;
				}
				paint_center (view_w, view_h / 3.0, foreground, selected, context);
				context.save ();
				selected = new bool[my_cards.length];
				for (size_t i = 0; i < my_cards.length; i++) {
					selected[i] = false;
				}
				context.translate (0, -view_h / 3.0);
				paint_hand (view_w, view_h / 3.0, my_cards, selected, context);
				context.restore ();
			}
		}

		public bool paint_game (double view_w,
								double view_h,
								Tarot.Game game,
								Cairo.Context context) {
			Tarot.Player view;
			Tarot.Card[] cards;
			bool has_view = (game.get_main_player (out view) == Tarot.GameError.OK);
			bool has_cards = (has_view && game.get_deal_of_alloc (view, out cards) == Tarot.GameError.OK);
			assert (has_view == has_cards);
			if (has_view && has_cards) {
				paint_game_as (view_w, view_h, game, view, context);
			}
			return has_view && has_cards;
		}
	}
}