/* window.vala
 *
 * Copyright 2019 Vivien Kraus
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace Tarot {

	private class EventQueue {

		// If there are more than that number of events, an assert is
		// raised!
		private Tarot.GameEvent m_data[256];
		private size_t m_head;
		private size_t m_n;
		private bool timer_running = false;
		private double next_delay = 0;

		public signal void new_event (Tarot.GameEvent e);
		public signal void catchup ();

		private void start_timer () {
			timer_running = true;
			double seconds = next_delay;
			next_delay = 1;
			switch (m_data[m_head].type ()) {
			case Tarot.GameEventT.SETUP_EVENT:
			case Tarot.GameEventT.DEAL_EVENT:
				next_delay = 0;
				break;
			case Tarot.GameEventT.DECL_EVENT:
				next_delay = 0.5;
				break;
			case Tarot.GameEventT.DOG_EVENT:
				next_delay = 3;
				break;
			default:
				next_delay = 1;
				break;
			}
			GLib.Timeout.add ((uint) (1000 * seconds), () => {
					timer_running = false;
					dequeue ();
					return false;
				});
		}

		private void dequeue () {
			assert (m_n > 0);
			new_event (m_data[m_head]);
			m_head ++;
			m_n --;
			m_head %= m_data.length;
			if (m_n > 0) {
				assert (!timer_running);
				start_timer ();
			} else {
				catchup ();
			}
		}

		public void enqueue (Tarot.GameEvent e) {
			assert (m_n < m_data.length);
			size_t i = (m_head + m_n) % m_data.length;
			m_data[i] = e;
			if (m_n++ == 0) {
				assert (!timer_running);
				start_timer ();
			}
		}
	}
	
	[GtkTemplate (ui = "/eu/planete_kraus/Tarot/window.ui")]
	public class Window : Gtk.ApplicationWindow {
		
		[GtkChild]
		private Tarot.Widget tarot_widget;
		
		private Tarot.Game my_game = new Tarot.Game.alloc ();
		private Tarot.ServerGame server;
		private Tarot.MctsGame?[] ais;
		private Tarot.Player myself = 0;
		private EventQueue m_queue = new EventQueue ();

		private void add_server_event (Tarot.GameEvent e) {
			if (my_game.add_event (e) != Tarot.GameError.OK) {
				// The server is desynchronized...  This is impossible
				// for now that we are in the same program.
				assert (false);
			}
			tarot_widget.game = my_game;
		}

		private void check_if_next () {
			tarot_widget.game = my_game;
			tarot_widget.dont_ask ();
			Tarot.Player next;
			Tarot.Bid mini;
			if (my_game.get_hint_bid (out next, out mini) == Tarot.GameError.OK
				&& next == myself) {
				stderr.printf ("Asking bid >= %s\n".printf (mini.to_string ()));
				tarot_widget.ask_bid (mini);
			}
			bool allowed;
			if (my_game.get_hint_decl (out next, out allowed) == Tarot.GameError.OK
				&& next == myself) {
				stderr.printf ("Asking declaration <= %s\n".printf (allowed.to_string ()));
				tarot_widget.ask_declaration (allowed);
			}
			Tarot.Number mini_call;
			if (my_game.get_hint_call (out next, out mini_call) == Tarot.GameError.OK
				&& next == myself) {
				stderr.printf ("Asking call <= %s\n".printf (mini_call.to_string ()));
				tarot_widget.ask_call (mini_call);
			}
			size_t n_discard;
			Tarot.Card[] prio;
			Tarot.Card[] additional;
			if (my_game.get_hint_discard_alloc (out next, out n_discard, out prio, out additional) == Tarot.GameError.OK
				&& next == myself) {
				stderr.printf ("Asking discard: %lu cards\n", n_discard);
				tarot_widget.ask_discard (n_discard, prio, additional);
			}
			size_t n_simple, n_double, n_triple;
			if (my_game.get_hint_handful_alloc (out next, out n_simple, out n_double, out n_triple, out prio, out additional) == Tarot.GameError.OK
				&& next == myself
				&& n_simple <= prio.length + additional.length) {
				stderr.printf ("Asking handful: %lu, %lu or %lu cards\n", n_simple, n_double, n_triple);
				tarot_widget.ask_handful (n_simple, n_double, n_triple, prio, additional);
			}
			Tarot.Card[] playable;
			if (my_game.get_hint_card_alloc (out next, out playable) == Tarot.GameError.OK
				&& next == myself) {
				stderr.printf ("Asking card");
				tarot_widget.ask_play (playable);
			}
		}

		construct {
			m_queue.new_event.connect (add_server_event);
			m_queue.catchup.connect (() => { check_if_next (); });
			tarot_widget.select_bid.connect ((bid) => {
					stderr.printf ("Selected bid %s\n", bid.to_string ());
					try {
						server.bid ((uint) bid, myself);
					} catch (ServerError err) {
						assert (false);
					}
				});
			tarot_widget.select_declaration.connect ((decl) => {
					stderr.printf ("Selected decl %s\n", decl.to_string ());
					try {
						server.declare (decl, myself);
					} catch (ServerError err) {
						assert (false);
					}
				});
			tarot_widget.select_call.connect ((card) => {
					stderr.printf ("Selected call %s\n", ((Tarot.Card) card).to_string_alloc ());
					try {
						server.call (card, myself);
					} catch (ServerError err) {
						assert (false);
					}
				});
			tarot_widget.select_discard.connect ((cards) => {
					stderr.printf ("Selected discard\n");
					try {
						server.discard (cards, myself);
					} catch (ServerError err) {
						assert (false);
					}
				});
			tarot_widget.select_handful.connect ((cards) => {
					stderr.printf ("Selected handful\n");
					try {
						server.handful (cards, myself);
					} catch (ServerError err) {
						assert (false);
					}
				});
			tarot_widget.select_card.connect ((card) => {
					stderr.printf ("Selected call %s\n", ((Tarot.Card) card).to_string_alloc ());
					try {
						server.play (card, myself);
					} catch (ServerError err) {
						assert (false);
					}
					tarot_widget.dont_ask ();
				});
		}

		public Window (Gtk.Application app, size_t n_players, bool with_call, MctsParams mp, int retry) {
			Object (application: app);
			try {
				server = new ServerGame (n_players, with_call);
			} catch (ServerError e) {
				stderr.printf (_ ("Error: could not set up the game: %s\n"), e.message);
				GLib.Process.exit (1);
			}
			ais = new Tarot.MctsGame?[n_players];
			for (size_t i = 0; i < n_players; i++) {
				ais[i] = null;
				if (i != myself) {
					ais[i] = new Tarot.MctsGame (new ServerSeat ((Tarot.Player) i, server));
					ais[i].seed = mp.seed;
					ais[i].n_iterations = mp.n_iterations;
					if (mp.parameter >= 0) {
						ais[i].parameter = mp.parameter;
					}
					if (mp.simulation_agreed >= 0) {
						ais[i].simulation_agreed = mp.simulation_agreed;
					}
					if (mp.simulation_random >= 0) {
						ais[i].simulation_random = mp.simulation_random;
					}
				}
			}
			string seed = mp.seed;
			while (!server.deal_random (seed)) {
				seed += ".";
			}
			server.event_setup.connect ((destination, n_players, with_call) => {
					var e = new Tarot.GameEvent.alloc_setup (n_players, with_call);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
					// The MCTS have received the event through their seats.
				});
			server.event_deal.connect ((destination, cards) => {
					var e = new Tarot.GameEvent.alloc_deal (destination, (Tarot.Card[]) cards);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_bid.connect ((destination, bid) => {
					var e = new Tarot.GameEvent.alloc_bid ((Tarot.Bid) bid);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_decl.connect ((destination, decl) => {
					var e = new Tarot.GameEvent.alloc_decl (decl);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_call.connect ((destination, call) => {
					var e = new Tarot.GameEvent.alloc_call (call);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_dog.connect ((destination, cards) => {
					var e = new Tarot.GameEvent.alloc_dog ((Tarot.Card[]) cards);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_discard.connect ((destination, cards) => {
					var e = new Tarot.GameEvent.alloc_discard ((Tarot.Card[]) cards);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_handful.connect ((destination, cards) => {
					var e = new Tarot.GameEvent.alloc_handful ((Tarot.Card[]) cards);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.event_play.connect ((destination, card) => {
					var e = new Tarot.GameEvent.alloc_card (card);
					if (destination == myself) {
						m_queue.enqueue (e);
					}
				});
			server.next.connect ((who) => {
					stderr.printf ("Next: %s\n", ((Tarot.Player) who).to_string_alloc ());
				});
			server.subscribe (myself);
		}
	}
}