// application.vala
//
// Copyright (C) 2019 Vivien Kraus <vivien@planete-kraus.eu>
// 
//  This program is free software: you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see
//  <http://www.gnu.org/licenses/>.

namespace Tarot {
	public class History<T>: GLib.Object {
		private T[] m_history;
		private int m_current;
		private int m_first;
		private int m_n;

		public int capacity {
			get {
				return m_history.length;
			}
			set {
				if (value < 1) {
					value = 1;
				}
				var buffer = new History<T>.with_capacity (value);
				for (int i = 0; i < m_n; i++) {
					var t = m_history[fold (i)];
					buffer.push (t);
				}
				m_history = buffer.m_history;
				m_current = buffer.m_current;
				m_first = buffer.m_first;
				m_n = buffer.m_n;
			}
		}

		public T? current {
			get {
				if (m_current < m_n) {
					return m_history[m_current];
				}
				return null;
			}
		}

		private int fold (int from_start) {
			assert (from_start < m_n);
			int pos = m_first + from_start;
			if (pos >= m_history.length) {
				pos -= m_history.length;
			}
			assert (pos >= 0 && pos < m_n);
			return pos;
		}

		private int unfold (int index) {
			assert (index >= 0 && index < m_n);
			int pos;
			if (index >= m_first) {
				pos = index - m_first;
			}
			else {
				pos = index + m_history.length - m_first;
			}
			assert (pos >= 0 && pos < m_n);
			return pos;
		}

		public bool undo () {
			int n = unfold (m_current);
			if (n != 0) {
				m_current = fold (n - 1);
			}
			return (n != 0);
		}

		public bool redo () {
			int n = unfold (m_current);
			if (n + 1 < m_n) {
				m_current = fold (n + 1);
			}
			return (n + 1 < m_n);
		}

		public void push (T data) {
			int n = 0;
			if (m_n > 0) {
				n = unfold (m_current);
				if (n + 1 < m_n) {
					n++;
					m_n = n; // Erase the redos
				} else if (m_n >= m_history.length) {
					// Forget the past
					m_first ++;
					if (m_first >= m_history.length) {
						m_first = 0;
					}
				} else {
					// There is room
					m_n ++;
				}
			} else {
				// First push ever
				m_n ++;
			}
			assert (n >= 0 && n < m_n);
			m_current = fold (n);
			m_history[m_current] = data;
		}

		public History () {
			this.with_capacity (1);
		}

		public History.with_capacity (int capacity) {
			m_history = new T[capacity];
			m_current = 0;
			m_first = 0;
			m_n = 0;
		}
	}
}