// main.vala
//
// Copyright (C) 2019 Vivien Kraus <vivien@planete-kraus.eu>
// 
//  This program is free software: you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see
//  <http://www.gnu.org/licenses/>.

[CCode (cname="RUNTIME_PREFIX_ENV")]
extern const string RUNTIME_PREFIX_ENV;

[CCode (cname="SYSCONFDIR")]
extern const string SYSCONFDIR;

[CCode (cname="LOCALEDIR")]
extern const string LOCALEDIR;

[CCode (cname="PACKAGE")]
extern const string PACKAGE;

[CCode (cname="PACKAGE_STRING")]
extern const string PACKAGE_STRING;

const string program_name = "tarot";

namespace Tarot {

	public struct MctsParams {
		unowned string? seed;
		size_t n_iterations;
		double parameter;
		double simulation_agreed;
		double simulation_random;
	}

	private static string read_input (GLib.FileStream input_source) {
		int c = 0;
		string ret = "";
		while ((c = input_source.getc ()) >= 0) {
			ret += string.nfill (1, (char) c);
		}
		return ret;
	}

	/// Return whether we have written some data.  Otherwise the file
	/// should be removed.
	private bool write_output (GLib.FileStream output, string? data) {
		if (data != null) {
			output.puts (data);
			return true;
		} else {
			return false;
		}
	}

	private GLib.OptionEntry create_option_none (string long_name, char short_name, ref bool? is_set, string documentation) {
		GLib.OptionEntry ret = GLib.OptionEntry ();
		ret.arg = OptionArg.NONE;
		ret.arg_data = &is_set;
		ret.arg_description = null;
		ret.description = documentation;
		ret.flags = 0;
		ret.long_name = long_name;
		ret.short_name = short_name;
		return ret;
	}

	private GLib.OptionEntry create_option_string (string long_name, char short_name, ref string? value, string documentation, string documentation_arg) {
		GLib.OptionEntry ret = GLib.OptionEntry ();
		ret.arg = OptionArg.STRING;
		ret.arg_data = &value;
		ret.arg_description = documentation_arg;
		ret.description = documentation;
		ret.flags = 0;
		ret.long_name = long_name;
		ret.short_name = short_name;
		return ret;
	}

	private GLib.OptionEntry create_option_int (string long_name, char short_name, ref int? value, string documentation, string documentation_arg) {
		GLib.OptionEntry ret = GLib.OptionEntry ();
		ret.arg = OptionArg.INT;
		ret.arg_data = &value;
		ret.arg_description = documentation_arg;
		ret.description = documentation;
		ret.flags = 0;
		ret.long_name = long_name;
		ret.short_name = short_name;
		return ret;
	}

	private GLib.OptionEntry create_option_float (string long_name, char short_name, ref double? value, string documentation, string documentation_arg) {
		GLib.OptionEntry ret = GLib.OptionEntry ();
		ret.arg = OptionArg.DOUBLE;
		ret.arg_data = &value;
		ret.arg_description = documentation_arg;
		ret.description = documentation;
		ret.flags = 0;
		ret.long_name = long_name;
		ret.short_name = short_name;
		return ret;
	}
	
	public static int main (string[] args) {
		int n_players = 4;
		bool call_allowed = false;
		string? seed = null;
		int retry = 0;
		int n_iterations = 100;
		double mcts_parameter = -1;
		double mcts_simulation_agreed = -1;
		double mcts_simulation_random = -1;
		string? output_resource = null;
		string? input_resource = null;
		int width = 640;
		int height = 480;
		bool print_version = false;
		bool has_status = false;
		bool has_new_game = false;
		string? has_deal_to = null;
		string? has_deal_all = null;
		bool has_deal_random = false;
		bool has_impute = false;
		bool has_mcts = false;
		bool has_mcts_tree = false;
		bool has_generate = false;
		string? has_bid = null;
		string? has_decl = null;
		string? has_call = null;
		string? has_dog = null;
		string? has_discard = null;
		string? has_handful = null;
		string? has_play = null;
		bool has_render = false;
		bool has_dbus_server = false;
		string default_server_bus_name = strdup ("eu.planete_kraus.Tarot");
		string server_bus_name = default_server_bus_name;
		string? has_dbus_mcts = null;
		bool has_anonymous_dbus = false;
		bool no_window = false;
		setup_environment (RUNTIME_PREFIX_ENV, SYSCONFDIR, LOCALEDIR);
		GLib.Intl.setlocale (LocaleCategory.ALL, "");
		GLib.Intl.bindtextdomain (PACKAGE, Environment.get_variable ("LOCALEDIR"));
		GLib.Intl.textdomain (PACKAGE);
		var options = new GLib.OptionEntry[0];
		options += create_option_none (_ ("new-game"), 0, ref has_new_game, _ ("Create a new game"));
		options += create_option_none (_ ("deal-random"), 0, ref has_deal_random, _ ("Deal the cards at random"));
		options += create_option_string (_ ("deal-to"), 0, ref has_deal_to, _ ("Deal cards to a specific player"), _ ("PN=CARD,CARD,..."));
		options += create_option_string (_ ("deal-all"), 0, ref has_deal_all, _ ("Deal the cards to all players at once"), _ ("OWNER1,OWNER2,..."));
		options += create_option_string (_ ("bid"), 0, ref has_bid, _ ("Add a bid to the game: 'pass', 'take', 'push', 'straight-keep', or 'double-keep'"), _ ("BID"));
		options += create_option_string (_ ("declare"), 0, ref has_decl, _ ("Add a slam declaration to the game"), _ ("'pass'/'slam'"));
		options += create_option_string (_ ("call"), 0, ref has_call, _ ("Call a card"), _ ("CARD"));
		options += create_option_string (_ ("dog"), 0, ref has_dog, _ ("Reveal the dog"), _ ("CARD,CARD..."));
		options += create_option_string (_ ("discard"), 0, ref has_discard, _ ("The discard is done"), _ ("'' or CARD,CARD..."));
		options += create_option_string (_ ("handful"), 0, ref has_handful, _ ("Show a handful"), _ ("'' or CARD,CARD..."));
		options += create_option_string (_ ("play"), 0, ref has_play, _ ("Play a card"), _ ("CARD"));
		options += create_option_none (_ ("status"), 0, ref has_status, _ ("Print the status of the game"));
		options += create_option_none (_ ("generate"), 0, ref has_generate, _ ("Generate an entire game with MCTS moves"));
		options += create_option_none (_ ("mcts"), 0, ref has_mcts, _ ("Play the next move with MCTS"));
		options += create_option_none (_ ("mcts-tree"), 0, ref has_mcts_tree, _ ("Print the MCTS tree for the next move"));
		options += create_option_none (_ ("impute"), 0, ref has_impute, _ ("Impute the missing cards"));
		options += create_option_none (_ ("render"), 0, ref has_render, _ ("Render a game as PNG"));
		options += create_option_none (_ ("dbus-server"), 0, ref has_dbus_server, _ ("Run as a DBus server"));
		options += create_option_string (_ ("dbus-mcts"), 0, ref has_dbus_mcts, _ ("Run the MCTS as a DBus client"), _ ("SEAT_PATH"));
		options += create_option_string (_ ("input"), 'i', ref input_resource, _ ("Input file to read (defaults to stdin)"), _ ("FILE"));
		options += create_option_string (_ ("output"), 'o', ref output_resource, _ ("Output file to write (defaults to stdout)"), _ ("FILE"));
		options += create_option_int (_ ("n-player"), 'n', ref n_players, _ ("Specify the number of players for a new game"), _ ("N"));
		options += create_option_none (_ ("call-allowed"), 'c', ref call_allowed, _ ("The new game can have a call"));
		options += create_option_string (_ ("seed"), 's', ref seed, _ ("The random seed for MCTS-related operations (defaults to 'default')"), _ ("NON-EMPTY"));
		options += create_option_int (_ ("retry"), 'r', ref retry, _ ("Number of times to retry an operation that may fail out of bad luck"), _ ("N"));
		options += create_option_int (_ ("n-iterations"), 'I', ref n_iterations, _ ("Number of iterations for the MCTS exploration (defaults to 100)"), _ ("N"));
		options += create_option_float (_ ("mcts-parameter"), 0, ref mcts_parameter, _ ("Parameter value for the MCTS algorithm.  Pass a negative value to use the maintainer's default"), _ ("VALUE"));
		options += create_option_float (_ ("mcts-simulation-agreed"), 0, ref mcts_simulation_agreed, _ ("Tendency for the MCTS to use an agreed simulation.  Pass a negative value to use the maintainer's default"), _ ("VALUE"));
		options += create_option_float (_ ("mcts-simulation-random"), 0, ref mcts_simulation_random, _ ("Tendency for the MCTS to use a random simulation.  Pass a negative value to use the maintainer's default"), _ ("VALUE"));
		options += create_option_none (_ ("no-window"), 0, ref no_window, _ ("Do not open a graphical window"));
		options += create_option_int (_ ("width"), 'w', ref width, _ ("PNG width (defaults to 640)"), _ ("N"));
		options += create_option_int (_ ("height"), 'h', ref height, _ ("PNG height (defaults to 480)"), _ ("N"));
		options += create_option_none (_ ("anonymous-dbus"), 0, ref has_anonymous_dbus, _ ("Set this option if you want to run the D-Bus service anonymously."));
		string doc = _ ("Override the DBus server owned name (default: %s)").printf (default_server_bus_name);
		options += create_option_string (_ ("dbus-server-name"), 0, ref server_bus_name, doc, _ ("DBUS_NAME"));
		options += create_option_none (_ ("version"), 'v', ref print_version, _ ("Print version information and exit"));
		GLib.OptionEntry final = { null };
		options += final;
		int ret = 0;
		try {
			var opt_context = new OptionContext (_ ("- Tarot command-line options"));
			opt_context.set_help_enabled (true);
			opt_context.add_main_entries (options, null);
			opt_context.parse (ref args);
		} catch (OptionError e) {
			stderr.printf ("Error: %s\n", e.message);
			GLib.Process.abort ();
		}
		MctsParams mparams = MctsParams ();
		mparams.seed = seed;
		mparams.n_iterations = n_iterations;
		mparams.parameter = mcts_parameter;
		mparams.simulation_agreed = mcts_simulation_agreed;
		mparams.simulation_random = mcts_simulation_random;
		string? actual_output = null;
		unowned GLib.FileStream input_stream = stdin;
		GLib.FileStream file_input_stream;
		if (input_resource != null) {
			file_input_stream = GLib.FileStream.open (input_resource, "r");
			input_stream = file_input_stream;
		}
		if (print_version) {
			ret = run_version (PACKAGE_STRING);
		} else if (has_status) {
			ret = run_status (read_input (input_stream), out actual_output);
		} else if (has_impute) {
			if (mparams.seed == null) {
				mparams.seed = "default";
			}
			ret = run_impute (read_input (input_stream), mparams.seed, retry, out actual_output);
		} else if (has_mcts) {
			if (mparams.seed == null) {
				mparams.seed = "default";
			}
			ret = run_mcts (read_input (input_stream), mparams, out actual_output);
		} else if (has_mcts_tree) {
			if (mparams.seed == null) {
				mparams.seed = "default";
			}
			ret = run_mcts_tree (read_input (input_stream), mparams, out actual_output);
		} else if (has_new_game) {
			ret = run_new_game (n_players, call_allowed, out actual_output);
		} else if (has_deal_to != null) {
			ret = run_deal_to (read_input (input_stream), has_deal_to, out actual_output);
		} else if (has_deal_all != null) {
			ret = run_deal_all (read_input (input_stream), has_deal_all, out actual_output);
		} else if (has_deal_random) {
			if (mparams.seed == null) {
				mparams.seed = "default";
			}
			ret = run_deal_random (read_input (input_stream), mparams.seed, retry, out actual_output);
		} else if (has_bid != null) {
			Tarot.Bid bid_arg;
			if (has_bid == _ ("pass")) {
				bid_arg = Tarot.Bid.PASS;
			} else if (has_bid == _ ("take")) {
				bid_arg = Tarot.Bid.TAKE;
			} else if (has_bid == _ ("push")) {
				bid_arg = Tarot.Bid.PUSH;
			} else if (has_bid == _ ("straight-keep")) {
				bid_arg = Tarot.Bid.STRAIGHT_KEEP;
			} else if (has_bid == _ ("double-keep")) {
				bid_arg = Tarot.Bid.DOUBLE_KEEP;
			} else {
				stderr.printf ("Error: '%s' is not a bid\n", has_bid);
				return 1;
			}
			ret = run_bid (read_input (input_stream), bid_arg, out actual_output);
		} else if (has_decl != null) {
			bool decl_arg;
			if (has_decl == _ ("pass")) {
				decl_arg = false;
			} else if (has_decl == _ ("slam")) {
				decl_arg = true;
			} else {
				stderr.printf ("Error: '%s' is not a slam declaration\n", has_decl);
				return 1;
			}
			ret = run_decl (read_input (input_stream), decl_arg, out actual_output);
		} else if (has_call != null) {
			ret = run_call (read_input (input_stream), has_call, out actual_output);
		} else if (has_dog != null) {
			ret = run_dog (read_input (input_stream), has_dog, out actual_output);
		} else if (has_discard != null) {
			ret = run_discard (read_input (input_stream), has_discard, out actual_output);
		} else if (has_handful != null) {
			ret = run_handful (read_input (input_stream), has_handful, out actual_output);
		} else if (has_play != null) {
			ret = run_play (read_input (input_stream), has_play, out actual_output);
		} else if (has_generate) {
			if (mparams.seed == null) {
				mparams.seed = "default";
			}
			ret = run_generate (n_players, call_allowed, mparams, retry, out actual_output);
		} else if (has_render) {
			var png_bytes = new char[0];
			ret = run_render (read_input (input_stream), width, height, out png_bytes);
			unowned GLib.FileStream str = stdout;
			GLib.FileStream file_str;
			if (output_resource != null) {
				file_str = GLib.FileStream.open (output_resource, "w");
				if (file_str == null) {
					stderr.printf ("Could not open file %s.\n", output_resource);
					GLib.Process.abort ();
				}
				str = file_str;
			}
			foreach (char c in png_bytes) {
				str.putc (c);
			}
		} else if (has_dbus_server) {
			ret = 0;
			string main_object_name = "/eu/planete_kraus/Tarot/Server/v1";
			if (!has_anonymous_dbus) {
				Bus.own_name (BusType.SESSION, server_bus_name, BusNameOwnerFlags.NONE,
							  ((conn) => {
								  try {
									  conn.register_object (main_object_name, new Server (conn));
								  } catch (IOError e) {
									  stderr.printf (_ ("Could not register the D-Bus server, sorry: %s.\n"), e.message);
									  GLib.Process.exit (1);
								  }
							  }));
			} else {
				try {
					var conn = Bus.get_sync (BusType.SESSION);
					stdout.printf ("%s\n", conn.unique_name);
					stdout.flush ();
					conn.register_object (main_object_name, new Server (conn));
				} catch (IOError e) {
					stderr.printf (_ ("Could not register the D-Bus server, sorry: %s.\n"), e.message);
					GLib.Process.exit (1);
				}
			}
			new MainLoop ().run ();
			return ret;
		} else if (has_dbus_mcts != null) {
			MctsGame mcts = null;
			Seat seat = null;
			var loop = new MainLoop ();
			try {
				seat = Bus.get_proxy_sync (BusType.SESSION, server_bus_name, has_dbus_mcts);
				mcts = new MctsGame (seat);
				if (mparams.seed != null) {
					mcts.seed = mparams.seed;
				}
				mcts.n_iterations = mparams.n_iterations;
				mcts.parameter = mparams.parameter;
				mcts.simulation_agreed = mparams.simulation_agreed;
				mcts.simulation_random = mparams.simulation_random;
				mcts.finished.connect ((score) => {
						stdout.printf ("%d\n", score);
						stdout.flush ();
						loop.quit ();
					});
			} catch (IOError e) {
				stderr.printf ("DBus error: %s\n", e.message);
				GLib.Process.exit (1);
			}
			loop.run ();
			return 0;
		} else {
			string seed_now = new GLib.DateTime.now_utc ().to_string ();
			if (mparams.seed == null) {
				mparams.seed = seed_now;
			}
			ret = run_interactive (n_players, call_allowed, mparams, retry, no_window, out actual_output);
			// Discard the output
			actual_output = null;
		}
		Tarot.quit ();
		if (ret == 0 && actual_output != null) {
			unowned GLib.FileStream str = stdout;
			GLib.FileStream file_str;
			if (output_resource != null) {
				file_str = GLib.FileStream.open (output_resource, "w");
				if (file_str == null) {
					stderr.printf ("Could not open file %s.\n", output_resource);
					GLib.Process.abort ();
				}
				str = file_str;
			}
			write_output (str, actual_output);
		}
		return ret;
	}
}