// pretty_status.vala
//
// Copyright (C) 2019 Vivien Kraus <vivien@planete-kraus.eu>
// 
//  This program is free software: you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see
//  <http://www.gnu.org/licenses/>.

namespace Tarot {
	public class PrettyStatus {
		// The pretty status uses "cells" in a table.  Each cell
		// should always have the same number of terminal columns.
		// TRANSLATORS: keep them 8 terminal columns.  You can use
		// unicode.
		private static string print_bid_cell (Tarot.Bid bid) {
			switch (bid) {
			case Tarot.Bid.PASS:
				return _ ("  pass  ");
			case Tarot.Bid.TAKE:
				return _ ("  take  ");
			case Tarot.Bid.PUSH:
				return _ ("  push  ");
			case Tarot.Bid.STRAIGHT_KEEP:
				return _ ("str. kp.");
			case Tarot.Bid.DOUBLE_KEEP:
				return _ ("dbl. kp.");
			default:
				assert_not_reached ();
			}
		}
		private static string print_decl_cell (bool decl) {
			if (decl) {
				return _ ("  slam! ");
			} else {
				return _ ("  pass  ");
			}
		}
		private static string print_player_cell (Tarot.Player player, bool taker, bool partner) {
			// The 8 columns are divided as 3 + 2 + 3, with a 3-column
			// marker left and right, to mark the taker and the
			// partner respectively.
			string marker_left = (taker ? _ ("** ") : "   ");
			string marker_right = (partner ? _ (" * ") : "   ");
			string middle = "??";
			if (player == 0) {
				middle = _ ("P1");
			} else if (player == 1) {
				middle = _ ("P2");
			} else if (player == 2) {
				middle = _ ("P3");
			} else if (player == 3) {
				middle = _ ("P4");
			} else if (player == 4) {
				middle = _ ("P5");
			}
			return marker_left + middle + marker_right;
		}
		private static string print_card_cell (Tarot.Card card, bool leader, bool taker, bool cnf, bool cno, bool cnu) {
			// The 8 columns are divided as 2 + 2 + 1 + 2 + 1, with 2
			// columns for the leader marker, 2 columns for the face,
			// 1 column for the suit, 2 columns for the trick taker
			// marker, and 1 column for the posterior information
			// (cannot follow, overtrump, undertrump).  Do NOT
			// translate the spaces, because we need them when
			// printing the call, dog, discard.
			string marker_left = (leader ? "->" : "  ");
			// The excuse can use all 3 columns
			string middle = "EXC";
			Tarot.Suit suit;
			Tarot.Number number;
			if (Tarot.decompose (card, out number, out suit) == 0) {
				string s_number = number.to_string ();
				if (suit != Tarot.Suit.TRUMPS) {
					switch (number) {
					case 1:
						s_number = _ (" A");
						break;
					case 11:
						s_number = _ (" J");
						break;
					case 12:
						s_number = _ (" C");
						break;
					case 13:
						s_number = _ (" Q");
						break;
					case 14:
						s_number = _ (" K");
						break;
					}
				}
				if (s_number.length < 2) {
					s_number = " " + s_number;
				}
				string s_suit = "?";
				switch (suit) {
				case Tarot.Suit.HEARTS:
					s_suit = "♥";
					break;
				case Tarot.Suit.CLUBS:
					s_suit = "♣";
					break;
				case Tarot.Suit.DIAMONDS:
					s_suit = "♦";
					break;
				case Tarot.Suit.SPADES:
					s_suit = "♠";
					break;
				case Tarot.Suit.TRUMPS:
					s_suit = "★";
					break;
				}
				middle = s_number + s_suit;
			}
			string marker_right = (taker? " *" : "  ");
			string marker_posterior = " ";
			if (cnu) {
				marker_posterior = "#";
			} else if (cno) {
				marker_posterior = "☹"; // Sad!
			} else if (cnf) {
				marker_posterior = "/";
			}
			return marker_left + middle + marker_right + marker_posterior;
		}
		private static string print_row (string[] items) {
			string separator = "│";
			if (items.length == 0) {
				return "";
			}
			string ret = separator;
			for (int i = 0; i < items.length; i++) {
				ret += items[i] + separator;
			}
			return ret;
		}
		private static string print_separation (string first_separation, string middle_separation, string end_separation, size_t n_columns) {
			string placeholder = "────────";
			if (n_columns == 0) {
				return "";
			} else {
				string ret = first_separation + placeholder;
				for (int i = 1; i < n_columns; i++) {
					ret += middle_separation + placeholder;
				}
				ret += end_separation;
				return ret + "\n";
			}
		}
		private static string print_top (size_t n_columns) {
			return print_separation ("╭", "┬", "╮", n_columns);
		}
		private static string print_middle (size_t n_columns) {
			return print_separation ("├", "┼", "┤", n_columns);
		}
		private static string print_bottom (size_t n_columns) {
			return print_separation ("╰", "┴", "╯", n_columns);
		}
		private static string print_table_header (size_t n_players, Tarot.Player taker, bool has_partner, Tarot.Player partner) {
			var cells = new string[n_players];
			string top = print_top (n_players);
			for (Tarot.Player i = 0; i < n_players; i++) {
				cells[i] = print_player_cell (i, (i == taker), (has_partner && i == partner));
			}
			return top + print_row (cells) + "\n";
		}
		private static string print_table_bids (size_t n_players, Tarot.Bid[] bids) {
			var cells = new string[n_players];
			for (Tarot.Player i = 0; i < bids.length; i++) {
				cells[i] = print_bid_cell (bids[i]);
			}
			for (Tarot.Player i = bids.length ; i < n_players; i++) {
				cells[i] = "        ";
			}
			return print_middle (n_players) + print_row (cells) + "\n";
		}
		private static string print_table_declarations (size_t n_players, bool[] declarations) {
			if (true in declarations || declarations.length != n_players) {
				var cells = new string[n_players];
				for (Tarot.Player i = 0; i < declarations.length; i++) {
					cells[i] = print_decl_cell (declarations[i]);
				}
				for (Tarot.Player i = declarations.length ; i < n_players; i++) {
					cells[i] = "        ";
				}
				return print_middle (n_players) + print_row (cells) + "\n" + print_bottom (n_players);
			}
			return print_bottom (n_players);
		}
		private static string print_card (Tarot.Card card) {
			string cell = print_card_cell (card, false, false, false, false, false);
			// Normally, the markers are all spaces so we can safely remove them
			int first = 0;
			int last = cell.length;
			while (first < cell.length && cell[first] == ' ') {
				first++;
			}
			while (last > first && cell[last - 1] == ' ') {
				last--;
			}
			return cell[first:last];
		}
		private static string print_call (Tarot.Card card) {
			return _ (" - Call: %s\n").printf (print_card (card));
		}
		private static string print_optional_call (Tarot.Card? card) {
			if (card == null) {
				return "";
			} else {
				return print_call (card);
			}
		}
		private static string print_card_array (Tarot.Card[] cards) {
			size_t sz = cards.length;
			var items = new string[sz];
			for (size_t i = 0; i < sz; i++) {
				items[i] = print_card (cards[i]);
			}
			if (sz == 0) {
				return "";
			} else {
				string ret = "%s".printf (items[0]);
				for (size_t i = 1; i + 1 < sz; i++) {
					ret += ", %s".printf (items[i]);
				}
				if (sz > 1) {
					ret += ", %s".printf (items[sz - 1]);
				}
				return ret;
			}
		}
		private static string print_dog (Tarot.Card[] cards) {
			return _ (" - Dog: %s\n").printf (print_card_array (cards));
		}
		private static string print_optional_dog (Tarot.Card[]? cards) {
			if (cards == null) {
				return "";
			} else {
				return print_dog (cards);
			}
		}
		private static string print_discard (Tarot.Card[] cards, bool full) {
			if (cards.length == 0) {
				return _ (" - The discard has been done.\n");
			} else if (full) {
				return _ (" - Discard: %s\n").printf (print_card_array (cards));
			} else {
				return _ (" - Some cards of the discard: %s\n").printf (print_card_array (cards));
			}
		}
		private static string print_optional_discard (Tarot.Card[]? cards, bool full) {
			if (cards == null) {
				return "";
			} else {
				return print_discard (cards, full);
			}
		}
		private static string print_player (Tarot.Player player) {
			string cell = print_player_cell (player, false, false);
			// Normally, the markers are all spaces, like for the card
			int first = 0;
			int last = cell.length;
			while (first < cell.length && cell[first] == ' ') {
				first++;
			}
			while (last > first && cell[last - 1] == ' ') {
				last--;
			}
			return cell[first:last];
		}
		private static string print_handful (Tarot.Player who, int size, Tarot.Card[] cards) {
			if (size != 0) {
				switch (size) {
				case 1:
					return (_ (" - Simple handful of %s: %s\n")
							.printf (print_player (who), print_card_array (cards)));
				case 2:
					return (_ (" - Double handful of %s: %s\n")
							.printf (print_player (who), print_card_array (cards)));
				case 3:
					return (_ (" - Triple handful of %s: %s\n")
							.printf (print_player (who), print_card_array (cards)));
				default:
					assert_not_reached ();
				}
			}
			return "";
		}
		private static string print_handfuls (Tarot.Game game) {
			string ret = "";
			size_t n_players = game.n_players ();
			for (Tarot.Player i = 0; i < n_players; i++) {
				int size;
				Tarot.Card[] handful_cards;
				if (game.get_handful_alloc (i, out size, out handful_cards) == Tarot.GameError.OK) {
					ret += print_handful (i, size, handful_cards);
				}
			}
			return ret;
		}
		private static string print_trick (size_t i_trick, Tarot.Game game) {
			size_t n_players = game.n_players ();
			var cells = new string[n_players];
			Tarot.Player leader, taker;
			for (Tarot.Player i = 0; i < n_players; i++) {
				cells[i] = "        ";
			}
			if (game.get_trick_leader (i_trick, out leader) == Tarot.GameError.OK) {
				game.get_trick_taker (i_trick, out taker);
				Tarot.GameCardAnalysis[] cards;
				if (game.get_trick_cards_alloc (i_trick, out cards) != Tarot.GameError.OK) {
					assert (false);
					cards = {};
				}
				for (size_t i = 0; i < cards.length; i++) {
					Tarot.Player p = (Tarot.Player) ((leader + i) % n_players);
					cells[p] = print_card_cell (cards[i].card, p == leader, p == taker,
												(bool) cards[i].cannot_follow,
												(bool) cards[i].cannot_overtrump,
												(bool) cards[i].cannot_undertrump);
				}
				return print_row (cells) + "\n";
			}
			return "";
		}
		private static string print_tricks (Tarot.Game game) {
			size_t current_trick;
			if (game.get_current_trick (out current_trick) != Tarot.GameError.OK) {
				return "";
			}
			var n_tricks = current_trick + 1;
			if (n_tricks > game.n_tricks ()) {
				n_tricks = game.n_tricks ();
			}
			Tarot.Player taker;
			Tarot.Player partner;
			bool has_partner = (game.get_partner (out partner) == Tarot.GameError.OK);
			game.get_taker (out taker);
			if (!has_partner) {
				partner = taker;
			}
			var n_players = game.n_players ();
			string ret = print_table_header (game.n_players (), taker, has_partner, partner) + print_middle (n_players);
			for (size_t i = 0; i < n_tricks; i++) {
				ret += print_trick (i, game);
			}
			return ret;
		}
		private static string print_score (int[] scores) {
			size_t n_players = scores.length;
			var cells = new string[n_players];
			for (size_t i = 0; i < n_players; i++) {
				cells[i] = " %+6d ".printf (scores[i]);
			}
			return print_row (cells) + "\n" + print_bottom (n_players);
		}
		private static string print_tricks_and_score (Tarot.Game game) {
			string ret = print_tricks (game);
			if (ret == "") {
				return "";
			}
			int[] scores;
			bool has_scores = (game.get_scores_alloc (out scores) == Tarot.GameError.OK);
			size_t n_players = game.n_players ();
			if (has_scores) {
				return ret + print_middle (n_players) + print_score (scores);
			}
			return ret + print_bottom (n_players);
		}
		private static string print_remaining (Tarot.Player who, Tarot.Game game) {
			Tarot.Card[] cards;
			bool has_cards = (game.get_cards_alloc (who, out cards) == Tarot.GameError.OK);
			if (has_cards && cards.length != 0) {
				return _ (" - Cards of %s: %s\n").printf (print_player (who), print_card_array (cards));
			}
			return "";
		}
		private static string print_remainings (Tarot.Game game) {
			var ret = "";
			size_t n_players = game.n_players ();
			for (Tarot.Player i = 0; i < n_players; i++) {
				ret += print_remaining (i, game);
			}
			return ret;
		}
		private static string print_hint_bid (Tarot.Game game) {
			Tarot.Player player;
			Tarot.Bid mini;
			bool has_hint = (game.get_hint_bid (out player, out mini) == Tarot.GameError.OK);
			if (!has_hint) {
				return "";
			}
			string player_name = print_player (player);
			switch (mini) {
			case Tarot.Bid.PASS:
			case Tarot.Bid.TAKE:
				return _ ("%s may bid.\n").printf (player_name);
			case Tarot.Bid.PUSH:
				return _ ("%s may pass, or at least push.\n").printf (player_name);
			case Tarot.Bid.STRAIGHT_KEEP:
				return _ ("%s may pass, straight or double keep.\n").printf (player_name);
			case Tarot.Bid.DOUBLE_KEEP:
				return _ ("%s mau pass, or double keep.\n").printf (player_name);
			default:
				assert_not_reached ();
			}
		}
		private static string print_hint_decl (Tarot.Game game) {
			Tarot.Player player;
			bool allowed;
			bool has_hint = (game.get_hint_decl (out player, out allowed) == Tarot.GameError.OK);
			if (!has_hint) {
				return "";
			}
			string player_name = print_player (player);
			if (allowed) {
				return _ ("%s may declare slam or pass.\n").printf (player_name);
			} else {
				return _ ("%s may not declare or pass.\n").printf (player_name);
			}
		}
		private static string print_hint_call (Tarot.Game game) {
			Tarot.Player player;
			Tarot.Number mini;
			bool has_hint = (game.get_hint_call (out player, out mini) == Tarot.GameError.OK);
			if (!has_hint) {
				return "";
			}
			string player_name = print_player (player);
			switch (mini) {
			case 14:
				return _ ("%s may call a king.\n").printf (player_name);
			case 13:
				return _ ("%s may call a queen or a king.\n").printf (player_name);
			case 12:
				return _ ("%s may call a queen, a knight or a king.\n").printf (player_name);
			case 11:
				return _ ("%s may call any face card.\n").printf (player_name);
			default:
				assert_not_reached ();
			}
		}
		private static string print_hint_discard (Tarot.Game game) {
			Tarot.Player player;
			size_t n_cards;
			Tarot.Card[] prio;
			Tarot.Card[] additional;
			bool has_hint = (game.get_hint_discard_alloc (out player, out n_cards, out prio, out additional) == Tarot.GameError.OK);
			if (!has_hint) {
				return "";
			}
			string player_name = print_player (player);
			if (prio.length == 0) {
				return _ ("%s should discard %lu cards\n").printf (player_name, n_cards);
			} else if (prio.length > n_cards) {
				return (_ ("%s should discard %lu cards among %s\n")
						.printf (player_name, n_cards, print_card_array (prio)));
			} else if (prio.length == n_cards) {
				return (_ ("%s should discard exactly these %lu cards: %s\n")
						.printf (player_name, n_cards, print_card_array (prio)));
			} else {
				return (_ ("%s should discard %s, and %lu cards among %s\n")
						.printf (player_name, print_card_array (prio),
								 n_cards - prio.length,
								 print_card_array (additional)));
			}
		}
		private static string print_hint_handful (Tarot.Game game) {
			Tarot.Player player;
			size_t n_simple, n_double, n_triple;
			Tarot.Card[] prio;
			Tarot.Card[] additional;
			bool has_hint = (game.get_hint_handful_alloc (out player, out n_simple, out n_double, out n_triple, out prio, out additional) == Tarot.GameError.OK);
			if (!has_hint) {
				return "";
			}
			string player_name = print_player (player);
			// Simple handful
			string ret = "";
			if (prio.length + additional.length >= n_simple) {
				if (prio.length > n_simple) {
					ret += (_ ("%s can show a simple handful with %lu cards among %s\n")
							.printf (player_name, n_simple, print_card_array (prio)));
				} else if (prio.length == n_simple) {
					ret += (_ ("%s can show a simple handful with these %lu cards: %s\n")
							.printf (player_name, n_simple, print_card_array (prio)));
				} else {
					ret += (_ ("%s can show a simple handful with %s, and %lu cards among %s\n")
							.printf (player_name, print_card_array (prio),
									 n_simple - prio.length,
									 print_card_array (additional)));
				}
			}
			if (prio.length + additional.length >= n_double) {
				if (prio.length > n_double) {
					ret += (_ ("%s can show a double handful with %lu cards among %s\n")
							.printf (player_name, n_double, print_card_array (prio)));
				} else if (prio.length == n_double) {
					ret += (_ ("%s can show a double handful with these %lu cards: %s\n")
							.printf (player_name, n_double, print_card_array (prio)));
				} else {
					ret += (_ ("%s can show a double handful with %s, and %lu cards among %s\n")
							.printf (player_name, print_card_array (prio),
									 n_double - prio.length,
									 print_card_array (additional)));
				}
			}
			if (prio.length + additional.length >= n_triple) {
				if (prio.length > n_triple) {
					ret += (_ ("%s can show a triple handful with %lu cards among %s\n")
							.printf (player_name, n_triple, print_card_array (prio)));
				} else if (prio.length == n_triple) {
					ret += (_ ("%s can show a triple handful with these %lu cards: %s\n")
							.printf (player_name, n_triple, print_card_array (prio)));
				} else {
					ret += (_ ("%s can show a triple handful with %s, and %lu cards among %s\n")
							.printf (player_name, print_card_array (prio),
									 n_triple - prio.length,
									 print_card_array (additional)));
				}
			}
			return ret;
		}
		private static string print_hint_card (Tarot.Game game) {
			Tarot.Player player;
			Tarot.Card[] playable;
			bool has_hint = (game.get_hint_card_alloc (out player, out playable) == Tarot.GameError.OK);
			if (!has_hint) {
				return "";
			}
			string player_name = print_player (player);
			if (playable.length == 0) {
				return _ ("%s may play a card\n").printf (player_name);
			} else if (playable.length == 1) {
				return _ ("%s may play %s\n").printf (player_name, print_card (playable[0]));
			} else {
				return _ ("%s may play a card among %s\n").printf (player_name, print_card_array (playable));
			}
		}
		public static string print (Tarot.Game game) {
			size_t n_players = game.n_players ();
			Tarot.Player taker;
			Tarot.Player partner;
			game.get_taker (out taker);
			bool has_partner = (game.get_partner (out partner) == Tarot.GameError.OK);
			if (!has_partner) {
				partner = taker;
			}
			Tarot.Bid[] bids;
			if (game.get_bids_alloc (out bids) != Tarot.GameError.OK) {
				bids = new Tarot.Bid[0];
			}
			bool[] decls;
			if (game.get_declarations_alloc (out decls) != Tarot.GameError.OK) {
				decls = new bool[0];
			}
			Tarot.Card? call = null;
			if (game.get_call (out call) != Tarot.GameError.OK) {
				call = null;
			}
			Tarot.Card[]? dog = null;
			if (game.get_dog_alloc (out dog) != Tarot.GameError.OK) {
				dog = null;
			}
			Tarot.Card[]? discard = null;
			bool discard_full = false;
			if (game.get_full_discard_alloc (out discard) != Tarot.GameError.OK) {
				if (game.get_public_discard_alloc (out discard) != Tarot.GameError.OK) {
					discard = null;
				} 
			} else {
				discard_full = true;
			}
			return (print_table_header (n_players, taker, has_partner, partner)
					+ print_table_bids (n_players, bids)
					+ print_table_declarations (n_players, decls)
					+ print_optional_call (call)
					+ print_optional_dog (dog)
					+ print_optional_discard (discard, discard_full)
					+ print_handfuls (game)
					+ print_tricks_and_score (game)
					+ print_remainings (game)
					+ print_hint_bid (game)
					+ print_hint_decl (game)
					+ print_hint_call (game)
					+ print_hint_discard (game)
					+ print_hint_handful (game)
					+ print_hint_card (game));
		}
	}
}