// solo.vala
//
// Copyright (C) 2019 Vivien Kraus <vivien@planete-kraus.eu>
// 
//  This program is free software: you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see
//  <http://www.gnu.org/licenses/>.

extern string? readline (string prompt);
extern void add_history (string command);

[CCode (cname="PACKAGE_VERSION")]
extern const string tarot_version;

namespace Tarot {
	public class Solo: GLib.Object {
		public static int run (size_t n_players, bool call, MctsParams mp, int retry, out string? output) {
			var app = new Tarot.Application ();
			var my_game = new Tarot.Game.alloc ();
			Tarot.Player myself = 0;
			output = null;
			app.broadcast_event.connect ((e) => {
					if (my_game.add_event (e) != Tarot.GameError.OK) {
						stderr.printf (_ ("Could not add an event of type %s to game %s.\n"),
									   e.type ().to_string (), my_game.save_to_xml_alloc ());
						assert (false);
					}
				});
			app.unicast_event.connect ((e, who) => {
					if (who == myself) {
						if (my_game.add_event (e) != Tarot.GameError.OK) {
							stderr.printf (_ ("Could not add an event of type %s to game %s.\n"),
										   e.type ().to_string (), my_game.save_to_xml_alloc ());
							assert (false);
						}
					}
				});
			app.reset.connect (() => {
					my_game = new Tarot.Game.alloc ();
				});
			if (!app.new_game (n_players, call)) {
				stderr.printf (_ ("Error: wrong number of players (%lu) or the call is not accepted\n"),
							   n_players);
				return 1;
			}
			if (!app.deal_random (mp.seed, retry)) {
				stderr.printf (_ ("Error: wrong seed '%s', try another one\n"), mp.seed);
				return 1;
			}
			bool eof = false;
			stdout.printf (_ ("tarot %s\nCopyright (C) 2019 Vivien Kraus\nThis program comes with ABSOLUTELY NO WARRANTY.\nThis is free software, and you are welcome to redistribute it under\nthe conditions of the GNU General Public License, version 3 or later.\nYou should have received a copy of the GNU General Public License\nalong with this program.  If not, see <http://www.gnu.org/licenses/>.\n"),
						   tarot_version);
			while (!eof && my_game.step () != Tarot.Step.END) {
				var ss = Tarot.PrettyStatus.print (my_game);
				Tarot.Player next;
				bool has_next = (my_game.get_next (out next) == Tarot.GameError.OK);
				stdout.printf ("%s\n", ss);
				assert (has_next);
				if (next != myself) {
					if (!app.mcts (mp)) {
						stderr.printf (_ ("Error: could not auto play.  Try with more iterations.\n"));
						return 1;
					}
				} else {
					string help_about_card_names = _ ("The cards are formed by a number followed by a suit indicator.  Examples are:\n - AH: the ace of hearts;\n - 10C: the ten of clubs;\n - JS: the jack of spades;\n - CD: the knight of diamonds;\n - QH: the queen of hearts;\n - KD: the knight of diamonds;\n - 12T: the twelve of trumps;\n - EXC: the excuse.\n");
					var step = my_game.step ();
					switch (step) {
					case Tarot.Step.BIDS:
						var line = readline (_ ("bid> "));
						if (line != null) {
							line = line.strip ();
							if (line == _ ("pass") && app.bid (Tarot.Bid.PASS)) {
								add_history (_ ("pass"));
							} else if (line == _ ("take") && app.bid (Tarot.Bid.TAKE)) {
								add_history (_ ("take"));
							} else if (line == _ ("push") && app.bid (Tarot.Bid.PUSH)) {
								add_history (_ ("push"));
							} else if (line == _ ("straight keep") && app.bid (Tarot.Bid.STRAIGHT_KEEP)) {
								add_history (_ ("straight keep"));
							} else if (line == _ ("double keep") && app.bid (Tarot.Bid.DOUBLE_KEEP)) {
								add_history (_ ("double keep"));
							} else {
								stderr.printf (_ ("The bids are 'pass', 'take', 'push', 'straight keep' and 'double keep', but you have to outbid.\n"));
							}
						} else {
							eof = true;
						}
						break;
					case Tarot.Step.DECLS:
						var line = readline (_ ("declare> "));
						if (line != null) {
							line = line.strip ();
							if (line == _ ("pass") && app.decl (false)) {
								add_history (_ ("pass"));
							} else if (line == _ ("slam") && app.decl (true)) {
								add_history (_ ("slam"));
							} else {
								stderr.printf (_ ("The declarations are 'pass' or 'slam'.\n"));
							}
						} else {
							eof = true;
						}
						break;
					case Tarot.Step.CALL:
						var line = readline (_ ("call> "));
						if (line != null) {
							line = line.strip ();
							var c = Tarot.string_to_card (line);
							if (c == ((Tarot.Card) (-1)) || !app.call (c)) {
								stderr.printf (_ ("You should call a face card.  Example: KH (king of hearts), QC (queen of clubs), CD (knight of diamonds), JS (jack of spades).  Not all cards are possible.\n"));
							} else {
								add_history (c.to_string_alloc ());
							}
						} else {
							eof = true;
						}
						break;
					case Tarot.Step.DISCARD:
						var line = readline (_ ("discard> "));
						if (line != null) {
							line = line.strip ();
							var split = line.split (" ");
							var split_filtered = new string[0];
							foreach (var c in split) {
								if (c != "") {
									split_filtered += c;
								}
							}
							split = split_filtered;
							var cards = new Tarot.Card[split.length];
							var ok = true;
							for (int i = 0; i < split.length; i++) {
								var c = Tarot.string_to_card (split[i]);
								if (c == ((Tarot.Card) (-1))) {
									stderr.printf (_ ("'%s' is not a valid card.  %s"), split[i], help_about_card_names);
									ok = false;
									break;
								}
								cards[i] = c;
								split[i] = c.to_string_alloc ();
							}
							if (!ok) {
								/* Pass */
							}
							else if (app.discard (cards)) {
								add_history (string.join (" ", split));
							} else {
								stderr.printf (_ ("This discard is not possible.\n"));
							}
						} else {
							eof = true;
						}
						break;
					case Tarot.Step.TRICKS:
						size_t current_trick;
						bool has_current_trick = (my_game.get_current_trick (out current_trick) == Tarot.GameError.OK);
						if (has_current_trick && current_trick == 0) {
							var line = readline (_ ("play / show> "));
							if (line != null) {
								line = line.strip ();
								var split = line.split (" ");
								if (split.length == 0) {
									stderr.printf (_ ("Enter multiple cards to show a handful, or one to play.  %s"), help_about_card_names);
								} else if (split.length == 1) {
									var card_to_play = Tarot.string_to_card (split[0]);
									if (card_to_play == ((Tarot.Card) (-1))
										|| !app.play (card_to_play)) {
										stderr.printf (_ ("'%s' is not a valid card to play.  %s"), split[0], help_about_card_names);
									} else {
										add_history (card_to_play.to_string_alloc ());
									}
								} else {
									var cards = new Tarot.Card[split.length];
									var ok = true;
									for (int i = 0; i < split.length; i++) {
										var c = Tarot.string_to_card (split[i]);
										if (c == ((Tarot.Card) (-1))) {
											stderr.printf (_ ("'%s' is not a valid card.  %s"), split[i], help_about_card_names);
											ok = false;
											break;
										}
										cards[i] = c;
										split[i] = c.to_string_alloc ();
									}
									if (!ok) {
										/* Pass */
									} else if (app.handful (cards)) {
										add_history (string.join (" ", split));
									} else {
										stderr.printf (_ ("This handful is not possible.\n"));
									}
								}
							} else {
								eof = true;
							}
						} else {
							var line = readline (_ ("play> "));
							if (line != null) {
								line = line.strip ();
								var card_to_play = Tarot.string_to_card (line);
								if (card_to_play == ((Tarot.Card) (-1))
									|| !app.play (card_to_play)) {
									stderr.printf (_ ("'%s' is not a valid card to play.  %s"), line, help_about_card_names);
								} else {
									add_history (card_to_play.to_string_alloc ());
								}
							} else {
								eof = true;
							}
						}
						break;
					default:
						GLib.Process.abort ();
					}
				}
			}
			var ss = Tarot.PrettyStatus.print (my_game);
			stdout.printf ("%s\n", ss);
			output = app.xml;
			return 0;
		}
	}
}