[CCode (cname="RUN_INTERACTIVE")]
extern int RUN_INTERACTIVE (size_t n_players, bool call, Tarot.MctsParams mp, size_t retry, bool no_window, out string? output);

namespace Tarot {
	private static Tarot.Card parse_one_card (string card) {
		unowned string end;
		Tarot.Card ret = Card.parse (card, out end);
		if (end == card) {
			stderr.printf (_ ("ERROR: expected a card, not '%s'.\n"),
						   card);
			GLib.Process.abort ();
		} else if (end != "") {
			stderr.printf (_ ("ERROR: trailing garbage after card %s in '%s': '%s'.\n"),
						   ret.to_string_alloc (),
						   card, end);
			GLib.Process.abort ();
		}
		return ret;
	}
	private static Tarot.Player parse_one_player (string player) {
		unowned string end;
		Tarot.Player ret = Player.parse (player, out end);
		if (end == player) {
			stderr.printf (_ ("ERROR: expected a player, not '%s'.\n"),
						   player);
			GLib.Process.abort ();
		} else if (end != "") {
			stderr.printf (_ ("ERROR: trailing garbage after player %s in '%s': '%s'.\n"),
						   ret.to_string_alloc (),
						   player, end);
			GLib.Process.abort ();
		}
		return ret;
	}
	private static Tarot.Card[] parse_card_array (string cards) {
		string[] atoms = cards.split (",");
		Tarot.Card[] ret = new Tarot.Card[0];
		foreach (unowned string a in atoms) {
			ret += parse_one_card (a);
		}
		return ret;
	}
	private static Tarot.Player[] parse_player_array (string players) {
		string[] atoms = players.split (",");
		Tarot.Player[] ret = new Tarot.Player[0];
		foreach (unowned string a in atoms) {
			ret += parse_one_player (a);
		}
		return ret;
	}
	private static Tarot.Card[] parse_named_card_array (string named, out Tarot.Player whom) {
		string[] atoms = named.split ("=");
		if (atoms.length != 2) {
			stderr.printf (_ ("ERROR: the argument should be of the form PN=C1,C2,..., not '%s'"),
						   named);
			GLib.Process.abort ();
		}
		whom = parse_one_player (atoms[0]);
		return parse_card_array (atoms[1]);
	}
	public static void setup_environment (string runtime_prefix_env, string sysconfdir, string localedir) {
		string runtime_prefix = "";
		if (runtime_prefix_env != "" && runtime_prefix_env != "no") {
			string? test = GLib.Environment.get_variable (runtime_prefix_env);
			if (test == null) {
				stderr.printf (_ ("ERROR: environment variable '%s' is not set.  Try reinstalling the program.\n"),
							   runtime_prefix_env);
				GLib.Process.abort ();
			} else {
				runtime_prefix = test;
			}
		}
		string absolute_sysconfdir = runtime_prefix + sysconfdir;
		string absolute_localedir = runtime_prefix + localedir;
		GLib.Environment.set_variable ("SYSCONFDIR", absolute_sysconfdir, true);
		GLib.Environment.set_variable ("LOCALEDIR", absolute_localedir, true);
		if (Tarot.init (absolute_localedir) != 0) {
			stderr.printf (_ ("ERROR: could not initialize libtarot.\n"));
			GLib.Process.abort ();
		}
	}
	public static int run_version (string package_string) {
		stdout.printf (_ ("%s (libtarot %s)\n\nCopyright © 2019 Vivien Kraus\nThis is free software; se the source for copying conditions.  There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
					   package_string,
					   Tarot.libtarot_version ());
		return 0;
	}
	public static int run_impute (string input, string seed, int retry, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.load_from_xml (input)
			|| !app.impute (seed, retry)) {
			stderr.printf (_ ("ERROR: this game is not valid.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_mcts (string input, MctsParams mp, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.load_from_xml (input)
			|| !app.mcts (mp)) {
			stderr.printf (_ ("ERROR: could not explore with MCTS.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_deal_random (string input, string seed, int retry, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.load_from_xml (input)
			|| !app.deal_random (seed, retry)) {
			stderr.printf (_ ("ERROR: could not use that random seed.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_generate (int n_players, bool with_call, MctsParams mp, int retry, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.generate (n_players, with_call, mp, retry)) {
			stderr.printf (_ ("ERROR: could not use that random seed.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_mcts_tree (string input, MctsParams mp, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (app.load_from_xml (input)) {
			var printed = app.mcts_tree_to_xml (mp);
			if (printed == null) {
				stderr.printf (_ ("ERROR: not enough iterations.\n"));
				return 1;
			}
			output = printed;
		}
		return 0;
	}
	public static int run_new_game (size_t n_players, bool with_call, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.new_game (n_players, with_call)) {
			stderr.printf (_ ("ERROR: wrong number of players / call combination.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_deal_to (string input, string str, out string? output) {
		var app = new Tarot.Application ();
		Tarot.Player whom;
		Tarot.Card[] cards = parse_named_card_array (str, out whom);
		output = null;
		if (!app.load_from_xml (input)
			|| !app.deal_to (whom, cards)) {
			stderr.printf (_ ("ERROR: could not use that random seed.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_deal_all (string input, string owners, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.load_from_xml (input)
			|| !app.deal_all (parse_player_array (owners))) {
			stderr.printf (_ ("ERROR: could not use that random seed.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_bid (string input, Tarot.Bid bid, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.load_from_xml (input)
			|| !app.bid (bid)) {
			stderr.printf (_ ("ERROR: invalid bid.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_decl (string input, bool decl, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.load_from_xml (input)
			|| !app.decl (decl)) {
			stderr.printf (_ ("ERROR: invalid slam declaration.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_call (string input, string card, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.load_from_xml (input) || !app.call (parse_one_card (card))) {
			stderr.printf (_ ("ERROR: invalid or unexpected call.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_status (string input, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.load_from_xml (input)) {
			stderr.printf (_ ("ERROR: invalid game.\n"));
			return 1;
		}
		output = Tarot.PrettyStatus.print (app.game);
		return 0;
	}
	public static int run_dog (string input, string cards, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.load_from_xml (input) || !app.dog (parse_card_array (cards))) {
			stderr.printf (_ ("ERROR: invalid or unexpected dog revelation.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_discard (string input, string cards, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.load_from_xml (input) || !app.discard (parse_card_array (cards))) {
			stderr.printf (_ ("ERROR: invalid discard.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_handful (string input, string cards, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.load_from_xml (input) || !app.handful (parse_card_array (cards))) {
			stderr.printf (_ ("ERROR: invalid handful.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_play (string input, string card, out string? output) {
		var app = new Tarot.Application ();
		output = null;
		if (!app.load_from_xml (input)
			|| !app.play (parse_one_card (card))) {
			stderr.printf (_ ("ERROR: invalid move.\n"));
			return 1;
		}
		output = app.xml;
		return 0;
	}
	public static int run_render (string input, int width, int height, out char[] output) {
		var game = new Tarot.Game.alloc ();
		output = new char[0];
		if (game.load_from_xml (input) != Tarot.GameError.OK) {
			stderr.printf (_ ("ERROR: this game is not valid.\n"));
			return 1;
		}
		var surface = new Cairo.ImageSurface (Cairo.Format.ARGB32, width, height);
		var context = new Cairo.Context (surface);
		context.translate (width / 2.0, height / 2.0);
		context.scale (1, -1);
		var painter = new TarotCairo.DefaultCardPainter ();
		if (!painter.paint_game (width, height, game, context)) {
			stderr.printf (_ ("ERROR: This game does not have a default player.\n"));
			return 1;
		}
		var total = new char[0];
		var status = surface.write_to_png_stream ((data) => {
				foreach (uchar u in data) {
					total += (char) u;
				}
				return Cairo.Status.SUCCESS;
			});
		if (status == Cairo.Status.SUCCESS) {
			output = total;
			return 0;
		}
		stderr.printf (_ ("Error: could not write PNG: %s\n"), status.to_string ());
		return 1;
	}
	public static int run_interactive (size_t n_players, bool call, MctsParams mp, int retry, bool no_window, out string? output) {
		string my_output = null;
		string valid_seed = mp.seed;
		(void) no_window;
		if (mp.seed == null) {
			valid_seed = "%ld".printf ((long) GLib.get_real_time ());
		}
		mp.seed = valid_seed;
		int ret = RUN_INTERACTIVE (n_players, call, mp, retry, no_window, out my_output);
		output = my_output;
		return ret;
	}
}